/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: arednsend.c
 * Author: Jed Marti
 * Description: Send an e-mail message to a server on the AREDN network or anywhere else for
 *   that matter. See meshmail.
 * Created: Tue Dec 10 11:54:30 2024
 * Revision History:
 *   Sun Dec 29 07:46:05 2024 Code review, delete SMS files.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include "mqtt.h"

// How much noise to generate.
extern uint8_t Noise;

// Log files, see log.c, used by LOG macro so hiddent.
extern char logMsg[];
extern FILE *LOGfh;

// If set, then delete the message after sending.
extern uint8_t deleteSMS;


void arednSend(char *rmt, int32_t portnum, char *fname)
/*     Send a file to a remote TCP device that is listening on a port. The destination needs to
 *   be running the meshmail server. The destination comes from the mesages to field i.e.
 *            >to<from ...
 *   Normally this is detected by the to field being an IPV4 address or the abbreviation of one.
 * Parameters:
 *   rmt: Name of remote device. This can be an IPV4 format number or DNS name.
 *   portnum: The port they should be listening on. If they're not there, we die quickly.
 *   fname: The file to send name.
 * External References:
 *   Noise: For debugging messages. 0 means none.
 *   logMsg[], LOGfh: used by LOG macro..
 * Returned values: none
 * Errors detected: no answer from remote, message too long.
 */
{
  int32_t r;
  uint8_t ibuf[MAXMSG];
  FILE    *fh;
  ssize_t rc;
  int32_t i, j;
  int32_t c;

  // Must be running there quickly.,
  if (Noise > 0)
    printf("%s:%s:%d Message to AREDN %s:%d\n", __FILE__, __func__, __LINE__, rmt, portnum);

  // Connect to remote. If we don't hear anything in 10 seconds, then punt. No server running.
  r = netConnect(rmt, portnum, 10);
  if (r == -1)
  {
    LOG("Remote AREDN mail service for %s on port %d no answer", rmt, portnum);
    fprintf(stderr, "%s:%s:%d Remote AREDN mail service for %s on port %d no answer\n",
            __FILE__, __func__, __LINE__, rmt, portnum);
    return;
  }

  // Open the file to send. If fails, close net connection.
  if ((fh = fopen(fname, "rb")) == NULL)
  {
    fprintf(stderr, "%s:%s:%d Can't open %s to send\n", __FILE__, __func__, __LINE__, fname);
    close(r);
    return;
  }
  
  // Read the message into our buffer. We limit the size since these come from a Meshtastic
  // device. If the message is too large, then close everything and quit.
  for (i = 0, c = fgetc(fh); c != EOF; c = fgetc(fh), ++i)
  {
    if (i >= MAXMSG)
    {
      LOG("Maximum Meshtastic message size exceeded");
      fprintf(stderr, "%s:%s:%d Maximum Meshtastic message size exceeded\n", __FILE__, __func__,
              __LINE__);
      fclose(fh);
      close(r);
      return;
    }
    ibuf[i] = c;
  }

  fclose(fh);

  // Delete file if so requested.
  if (deleteSMS)
  {
    sprintf(logMsg, "rm %s", fname);
    if (system(logMsg) != 0)
    {
      fprintf(stderr, "%s:%s:%d Can't delete message file %s because %s\n", __FILE__, __func__,
              __LINE__, fname, strerror(errno));
      LOG("Can't delete %s because %s", fname, strerror(errno));
    }
  }
  
  if (Noise > 1) printf("%s:%s:%d Sending %d bytes\n", __FILE__, __func__, __LINE__, i);

  // TCP/IP may require multiple writes but this is unlikely unless the net has a very small
  // MTU (< 256 bytes).
  for (j = 0;;)
  {
    rc = write(r, &ibuf[j], i - j);
    if (rc == (i - j)) break;
    j += rc;
  }
  close(r);

  // Log that we sent but don't save message contents.
  LOG("Sent message to %s:%d", rmt, portnum);
}



// EOF
