/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: clone.c
 * Author: Jed Marti
 * Description: Control of object cloning. We clone structures that need to be kept around so that
 *   the others can be garbage collected.
 * Created: Sat Dec 14 10:29:39 2024
 * Revision History:
 *   Sun Dec 29 07:56:04 2024 - Code review.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>

#include "mqtt.h"
#include "protobuf.h"

// Make some noise at high settings.
extern uint8_t Noise;

// Log files, see log.c, used by LOG macro
extern char logMsg[];
extern FILE *LOGfh;

// The table of clone functions by type.
cloneFn cloneFns[MAXTYPES];


static Generic *noClone(Generic *g)
/*     A default clone function for objects we don't know about. Causes a halt.
 * Parameters:
 *   g: Bad Generic type.
 * External References:
 *   LOGfh, logMsg: For logging the error.
 * Returned values: none
 * Errors detected: this is the end of an error chain.
 */
{
  LOG("Cloning invalid type descriptor 0x%02X", __descriptor(g));
  fprintf(stderr, "%s:%s:%d Can't clone 0x%02X\n", __FILE__, __func__, __LINE__, __descriptor(g));
  exit(-1);
}



Generic *clone(Generic *g)
/*     Clone a Generic structure. Verify it's OK, then call the correct function.
 * Parameters:
 *   g: Generic structure to clone.
 * External References:
 *   MAXTYPES: constant, how many Generic types are threre.
 *   LOGfh, logMsg: For logging an error.
 * Returned values: a copy of the structure (there is only one NULL).
 * Errors detected: Bad descriptor.
 */
{
  if (Noise > 5)
    printf("%s:%s:%d Cloning %s", __FILE__, __func__, __LINE__, typeName(g));

  if (g == NULL) return NULL;
  if (__descriptor(g) >= MAXTYPES)
  {
    LOG("Cloning invalid type descriptor 0x%02X", __descriptor(g));
    fprintf(stderr, "%s:%s:%d Can't clone 0x%02X\n", __FILE__, __func__, __LINE__, __descriptor(g));
    exit(-1);
  }
  return (cloneFns[__descriptor(g)])(g);
}



void initClone(void)
/*     Initialize the admittedly small table with error clones. The actual cloning gets done near
 *   the object definitions. Each type fills in its clone function during initialization. If you add
 *   a type make sure you add its clone function during its initialization.
 * Parameters: none
 * External References:
 *   cloneFns[]: table of functions by descriptor.
 * Returned values: none
 * Errors detected: none
 */
{
  int32_t i;
  for (i = 0; i < MAXTYPES; ++i) cloneFns[i] = noClone;
}



// EOF
