/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: ocmdline.c
 * Author: Jed Marti
 * Description: Parse command line options.
 * Created: Tue Dec  3 08:12:03 2024
 * Revision History:
 *   Sun Dec 29 07:58:24 2024 - code review, add deleteSMS files option (normally off).
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>

#include "mqtt.h"

// Portnumber to listen on.
int32_t portNumber = DEFAULT_PORTNUM;

// Noise level
uint8_t Noise = 0;

// Abbreviation file name. We look here for real world mail addresses for the abbreviations used
// by Meshtastic senders so they don't exceed message sizes by entering long e-mail addresses.
char *abbreviationFileName = "abbrevs";

// AREDN port number. A destination name of an IPV4 number indicates a simple mail server on the
// far end waiting for a message (see meshmail). This is the port number it's listening on.
int32_t AREDNPortnum = DEFAULT_AREDN;

// Packet tracing. 
uint8_t packetTr = 0;
char    *packetDir = TRACEDIR;

// SMS directory. This is where the Meshtastic mail is stored before being shipped off to the
// internet or the meshmail server. Typically this would be /tmp/. 
char    *SMSDir = SMSDIR;

// Delete after sending. Should Meshtastic e-mails be deleted after sending? Typically not. This
// can be set by the -d/-delete option.
uint8_t deleteSMS = 0;

// The log file.
char    *LOGFileName = LOGNAME;
FILE    *LOGfh;

// Start input server (experimental.
uint8_t startSTDIN = 0;

// Keep alive at startup extra time for watchdog with default.
uint32_t keepAliveStartup = KEEPALIVESTARTUP;

// Mail channel name.
char *mailChannel = ARESCHANNEL;



static int32_t nextarg(int32_t i, int32_t argc, const char *emsg)
/*     Repeated code. Move to next argument, make sure it's there.
 * Parameters:
 *   argc: Number of command line arguments.
 *   argv[]: Command line arguments as strings.
 *   emsg: What we're looking for in case it isn't there.
 * External References: none
 * Returned values: index of the requested argument.
 * Errors detected: Fell off end.
 */
{
  if (++i >= argc)
  {
    fprintf(stderr, "Looking for %s command line argument but not found\n", emsg);
    exit(-1);
  }
  return i;
}



void cmdline(int32_t argc, char *argv[])
/*      Scan the command line and modify the parameters listed (if any).
 * Parameters:
 *  argc: Count of arguments including program name.
 *  argv[]: argc count of the arguments as strings.
 * External References:
 *   portNumber: Port number we're listening on for MQTT interface.
 *   Noise: Noise level for tracing, 0=none, 5=garrulous.
 *   abbrefiationFile: File name to look for abbreviations in.
 *   AREDNPortnum: Port number to pass messages on to AREDN mesh network.
 *   packetDir: dump packets, where to put them.
 *   packetTr: Trace packets (dump in C).
 *   deleteSMS: Set true to delete mail after sending.
 * Returned values: none
 * Errors detected: bad values, missing arguments.
 */
{
  int32_t i;

  for (i = 1; i < argc; ++i)
  {

    // Port number.
    if (strcasecmp("-portnum", argv[i]) == 0 ||
        strcasecmp("-p", argv[i]) == 0)
    {
      i = nextarg(i, argc, "-portnum integer");
      portNumber = atoi(argv[i]);
      if (portNumber < 1024)
        fprintf(stderr, "Port number %d requires root priveleges\n", portNumber);
    }

    // Noise level. 0=quiet, 5 = everything.
    else if (strcasecmp("-noise", argv[i]) == 0 ||
             strcasecmp("-n", argv[i]) == 0)
    {
      i = nextarg(i, argc, "-noise integer");
      Noise = (uint8_t) atoi(argv[i]);
    }

    // Address file name. - means there isn't one.,
    else if (strcasecmp("-addresses", argv[i]) == 0 ||
             strcasecmp("-a", argv[i]) == 0)
    {
      i = nextarg(i, argc, "-abbreviations/-a filename");
      if (strcmp(argv[i], "-") == 0) abbreviationFileName = NULL;
      else  abbreviationFileName = argv[i];
    }

    // Extra time at startup.
    else if (strcasecmp("-k", argv[i]) == 0 ||
             strcasecmp("-keep", argv[i]) == 0)
    {
      i = nextarg(i, argc, "-k/-keep");
      keepAliveStartup = atoi(argv[i]);
    }
    
    // If file name is -, then no log file.
    else if (strcmp("-L", argv[i]) == 0 ||
             strcasecmp("-log", argv[i]) == 0)
    {
      i = nextarg(i, argc, "-L/-log dir");
      if (strcmp(argv[i], "-") == 0) LOGFileName = NULL;
      else LOGFileName = argv[i];
    }

    // The port number the AREDN mail server will sit.
    else if ((strcasecmp("-arednPortnum", argv[i]) == 0) ||
             (strcasecmp("-ap", argv[i]) == 0))
    {
      i = nextarg(i, argc, "-arednPortnum/-ap");
      AREDNPortnum = atoi(argv[i]);
      if (AREDNPortnum < 1024)
        fprintf(stderr, "Portnumber %d < 1024 requires root access\n", AREDNPortnum);
    }

    // Where to put SMS messages until sent.
    else if (strcasecmp("-smsdir", argv[i]) == 0 ||
             strcasecmp("-s", argv[i]) == 0)
    {
      i = nextarg(i, argc, "-smsdir/-s");
      SMSDir = argv[i];
    }

    // Turn stdin server on (defaults off).
    else if (strcasecmp("-m", argv[i]) == 0 ||
             strcasecmp("-mail", argv[i]) == 0) startSTDIN = 1;

    // Channel name for mail.
    else if (strcasecmp("-c", argv[i]) == 0 ||
             strcasecmp("-channel", argv[i]) == 0)
    {
      i = nextarg(i, argc, "-channel/-c");
      mailChannel = argv[i];
    }

    // Directory for packet tracing.
    else if (strcasecmp("-td", argv[i]) == 0)
    {
      i = nextarg(i, argc, "-td string");
      packetDir = argv[i];
    }

    // Save packets for analysis.
    else if (strcasecmp("-tr", argv[i]) == 0) packetTr = 1;

    // Do delete Meshtastic e-mails after sending.
    else if (strcasecmp("-d", argv[i]) == 0 ||
             strcasecmp("-delete", argv[i]) == 0) deleteSMS = 1;
         
    // help.
    else if (strcasecmp("-help", argv[i]) == 0 ||
             strcasecmp("-h", argv[i]) == 0)
    {
      fprintf(stderr, "\
Options:\n\
  -a  -addresses     filename  Where address abbreviations are found, - means none (%s)\n\
  -ap -arednPortnum  integer   AREDN port number %d.\n\
  -c  -channel       string    Channel name for mail(%s)\n\
  -d  -delete                  Delete e-mails after sending\n\
  -k  -keep          integer   Extra time at startup for watchdog (%u)\n\
  -L  -log           dirname   Where to put the log file (%s)\n\
  -m  -mail                    Start the stdin server\n\
  -n  -noise         integer   Set the output noise level (%d) 0-5\n\
  -p  -portnum       integer   Set listen port number (%d)\n\
  -s  -smsdir        dirname   Directory name to put SMS temporaries (%s)\n\
  -td                dirname   Directory for packet trace (%s)\n\
  -tr                          Turn on packet tracing\n\
\nYou must enable Uplink and Downlink in the Meshtastic channel settings.\n\
You may also need to remove a default gateway.",
              abbreviationFileName, AREDNPortnum, mailChannel, keepAliveStartup, LOGFileName, Noise,
              portNumber, SMSDir, packetDir);
      exit(0);
    }

    else
    {
      fprintf(stderr, "%s:%s:%d Option %s unknown\n", __FILE__, __func__, __LINE__, argv[i]);
      exit(-1);
    }
  }

}


// EOF
