/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: connect.c
 * Author: Jed Marti
 * Description: Work MQTT CONNECT packet. Retrieve and parse a message.
 * Created: Mon Dec  2 06:34:52 2024
 * Revision History:
 *   Sun Dec 29 08:16:40 2024 - code review.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include "mqtt.h"

// Keep alive - we'll get a keep alive message every so often. If we don't, then the watchdog
// at the top level will reset the connection.,
uint32_t keepAlive = 0;

// Network file handle - this is the connection to the meshtastic device which is either WiFi
// or ethernet. We do not do Bluetooth.
extern int32_t FD;

// Output noise level.
extern uint8_t Noise;

// Connect acknowledgement packet. Sent in response to the connection request.
static uint8_t connAck[] = { MQTT_CONNACK << 4, 2, 0, 0};

// To simulate getting a byte at a time.
extern int32_t packetIndex;

// Downlink status - used to indicate we have a connection so we can download stuff to the
// Meshtastic which doesn't seem to work.
extern uint8_t downStatus;

// Log files, see log.c, used by LOG macro
extern char logMsg[];
extern FILE *LOGfh;

// Our client ID as a number - this gets used in variuos places to indicate which device we're
// talking to and to pass messages along.
uint32_t clientID;



void mqttConnect(uint8_t pt, uint8_t *pkt)
/*     We've received a connection request from a Meshtastic device. Parse the MQTT packet and
 *   save whatever information we need for new messages.
 * Parameters:
 *   pt: The packet descriptor. Should be 0x01 (from caller). Ignored.
 *   pkt: The bytes from the connection message.
 * External References:
 *   Noise: How much noise to generate: 0=none, 1-? more and more.
 *   downStatus: Set to indicate downlink is established.
 *   packetIndex: Pointer to receive buffer.
 * Returned values: none
 * Errors detected: bad protocol, write fails.
 */
{
  uint16_t sz, i;
  int32_t  p;
  char     str[MAXMSG];
  uint8_t  plevel;
  uint8_t  pflags;
  ssize_t  rc;
  char     c;
  char     username[MAXNAMESIZE];
  char     password[MAXNAMESIZE];
  
  if (Noise > 0) printf("CONNECT\n");

  // Get protocol size and protocol - this is a two byte MQTT size, not protoBuf VARINT.
  // If this isn't an MQTT message, then we quit, somebody trying to break in.
  sz = pget16(pkt);
  if (sz >= MAXMSG)
  {
    fprintf(stderr, "%s:%s:%d Protocol size %u in error. Quitting\n", __FILE__, __func__, __LINE__,
            sz);
    LOG("Protocol size %u way too large. Quitting\n", sz);
    exit(-1);
  }
  for (i = 0; i < sz; ++i) str[i] = pgetch(pkt);
  str[i] = 0;
  if (strcmp(str, "MQTT") != 0)
  {
    fprintf(stderr, "%s:%s:%d '%s' non standard protocol\n", __FILE__, __func__, __LINE__, str);
    LOG("%s is non-standard protocol", str);
    exit(-1);
  }

  // Get protocol level, must be 4 (3.1.1)
  plevel = pgetch(pkt);
  if (plevel != 4)
  {
    fprintf(stderr, "%s:%s:%d Protocol %u illegal, quitting\n", __FILE__, __func__,
            __LINE__, plevel);
    LOG("Protocol level %u illegal, probably MQTT 5.0, quitting", plevel);
    exit(-1);
  }

  // Connect flags.
  pflags = pgetch(pkt);

  // Keep alive flags.
  keepAlive = pget16(pkt);
  if (Noise > 1)
    printf("  %s:%s:%d Keep alive %u seconds\n", __FILE__, __func__, __LINE__, keepAlive);
  
  // Client identification (Meshtastic MAC last 32 bits) as an ASCII character set.
  sz = pget16(pkt);
  if (sz > 8)
  {
    fprintf(stderr, "%s:%s:%d Client id size %d is too large.\n", __FILE__, __func__, __LINE__, sz);
    LOG("MQTT client id size %u more than 8 characters", sz);
  }
  clientID = 0;
  for (i = 0; i < sz; ++i)
  {
    c = pgetch(pkt);
    if (c >= '0' && c <= '9') clientID = (clientID << 4) + c - '0';
    else if (c >= 'a' && c <= 'f') clientID = (clientID << 4) + (c - 'a') + 10;
    else if (c >= 'A' && c <= 'F') clientID = (clientID << 4) + (c - 'A') + 10;
  }
      
  if (Noise > 0)
    printf("  %s:%s:%d Client id: !0x%0x\n", __FILE__, __func__, __LINE__, clientID);

  // The user name. Complicating by somebody trying to break the user name size limit.
  if (pflags & 0x80)
  {
    sz = pget16(pkt);
    if (sz >= MAXNAMESIZE)
    {
      fprintf(stderr, "%s:%s:%d User name size %u too large, truncating\n", __FILE__, __func__,
              __LINE__, sz);
      LOG("User name size %u too large, truncating", sz);
    }

    // Now read the name and truncate if necessary.
    if (Noise > 0)
    {
      printf("  %s:%s:%d User name: ", __FILE__, __func__, __LINE__);
      for (i = 0; i < sz; ++i)
        if (i < MAXNAMESIZE - 1) printf("%c", username[i] = pgetch(pkt));
        else printf("%c", pgetch(pkt));
      printf("\n");
    }
    else
      for (i = 0; i < sz; ++i)
        if (i < MAXNAMESIZE - 1) username[i] = pgetch(pkt);
        else pgetch(pkt);

    if (sz < MAXNAMESIZE - 1) username[sz] = 0;
    else username[MAXNAMESIZE - 1] = 0;
  }
  else username[0] = 0;

  // User password. If the password is too large, something is really wrong. Time to quit.
  if (pflags & 0x40)
  {
    sz = pget16(pkt);
    if (sz >= MAXNAMESIZE)
    {
      fprintf(stderr, "%s:%s:%d Password size %u too large, quitting\n", __FILE__, __func__,
              __LINE__, sz);
      LOG("Password size %u too large, quitting", sz);
      exit(-1);
    }
    
    if (Noise > 0)
    {
      printf("  %s:%s:%d Password: ", __FILE__, __func__, __LINE__);
      for (i = 0; i < sz; ++i) printf("%c", password[i] = pgetch(pkt));
      printf("\n");
    }
    else for (i = 0; i < sz; ++i) password[i] = pgetch(pkt);
    password[sz] = 0;
  }
  else password[0] = 0;

  // Log the connection.
  LOG("CONNECT from !%08x user '%s' password '%s' keep alive %d protocol %d",
      clientID, username, password, keepAlive, plevel);
  
  // Send the MQTT response packet. 
  rc = write(FD, connAck, sizeof(connAck));
  if (rc != sizeof(connAck))
  {
    fprintf(stderr, "%s:%s:%d CONNACK response write failed because %s\n",
            __FILE__, __func__, __LINE__, strerror(errno));
    LOG("CONNACK response write failed because %s", strerror(errno));
    exit(-1);
  }

  // The uplink established.
  //  downStatus = downUpConnected;
}



// EOF
