/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: gc.c
 * Author: Jed Marti
 * Description: Collecting garbage by type. Requires program call, not automatic. Should be called
 *   between message handling to remove previous message and at other times. Strings rely on the
 *   OS gc which isn't all that great and will generate a lot of memory fragmentation over time.
 *   Best to reset the system every few days if there are lots of messages.
 * Created: Fri Dec  6 11:24:44 2024
 * Revision History:
 *   Sun Dec 29 11:08:15 2024 - Code review.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>

#include "protobuf.h"

// How much noise to generate.
extern uint8_t Noise;

// Table to call for various types. If you add a new type, you must have a GC function for it and
// add it to this table at startup. The function must have one argument which is guaranteed to be
// a Generic of your new type. Typically these are added to a free list to be reused if they have
// a static size. If not, you have to rely on 'free'.
gcFn gcFns[MAXTYPES];

// Type namess.
static char *tNames[] = {"Int", "Double", "String", "Pair"};
static char *charNULL = "NULL";
char *charHex = "0x02";


char *typeName(Generic *g)
/*     Return a type name to be used for debugging.
 * Parameters:
 *  g: Any Generic, even illegal ones or NULL.
 * External References:
 *   charNULL: characters for NULL.
 *   MAXTYPES: constant (see protobuf.h).
 *   charHex: Characters for illegal descriptor.
 *   tNames[]: Names of known types.
 * Returned values: a static name. Do not GC this name.
 * Errors detected: Illegal descriptors tagged with their value.
 */
{
  if (g == NULL) return charNULL;
  if (__descriptor(g) >= MAXTYPES)
  {
    sprintf(charHex, "0x%02X", __descriptor(g));
    return charHex;
  }
  return tNames[__descriptor(g)];
}



void noSpace(const char *msg, const char* fname, const char *func, int32_t lno)
/*     Common function for malloc fails. Generated by the outOfSpace macro which fills in
 *   the caller information.
 * Parameters:
 *   msg: Message of what caused the crash.
 *   fname: Normally __FILE__ from macro.
 *   func: Normally __func__ from macro.
 *   lno: Normally __LINE__ from macro
 * External References: none
 * Returned values: none, does not return.
 * Errors detected: none
 */
{
  fprintf(stderr, "%s:%s:%d Out of space %s\n", fname, func, lno, msg);
  exit(-1);
}



static void illegalType(Generic *g)
/*     Trying to GC an unknown descriptor lands you here.
 * Parameters:
 *   g: Generic with bad descriptor.
 * External References: none
 * Returned values: does not return.
 * Errors detected: none, end of error chain.
 */
{
  fprintf(stderr, "GC illegal type 0x%02X\n", __descriptor(g));
  exit(-1);
}



void gc(Generic *g)
/*     Free up space for a Generic type. This can be recursive for structures.
 * Parameters:
 *   g: Generic with bad descriptor.
 * External References:
 *   gcFns[]: table of functions to call for each type.
 * Returned values: none
 * Errors detected: Bad descriptor causes halk.
 */
{
  if (Noise > 5)
    printf("%s:%s:%d GC starts on %s", __FILE__, __func__, __LINE__, typeName(g));
  if (g == NULL) return;
  if (__descriptor(g) > MAXTYPES) illegalType(g);
  (gcFns[__descriptor(g)])(g);
  if (Noise > 5)
    printf("%s:%s:%d GC complete", __FILE__, __func__, __LINE__);
}



void initGC(void)
/*     Initialize the GC table to all illegal. When new types show up, they should override
 *   the illegal. This must be called before any descriptors do their fill in.
 * Parameters: none
 * External References:
 *   gcFns[]: Table by type descriptor.
 *   MAXTYPES: size of table.
 * Returned values: none
 * Errors Detected: none
 */
{
  int32_t i;
  for (i = 0; i < MAXTYPES; ++i) gcFns[i] = illegalType;
}



// EOF
