/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: log.c
 * Author: Jed Marti
 * Description: Log stuff into whereever log files go. 
 * Created: Sun Dec 15 12:54:01 2024
 * Revision History:
 *   Sun Dec 29 15:17:33 2024 - Code review.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>

#include "mqtt.h"

// Global used by the LOG macro. Enough space for a short essay.
char logMsg[MAXPATHLEN * 2];

// The log file. See cmdline.c for changing values and mqtt.h for defaults. The file handle is
// opened at startup and kept open. Each write to the log is flushed so when the system crashes
// or power goes off the log will be there. The file name has the date it started in so it
// never gets overridden.
extern char *LOGFileName;
extern FILE *LOGfh;


void iLog(void)
/*     The internal logging function. The LOG macro (see mqtt.h) lets you use this like printf
 *   by doing an snprintf to the global and then writing it. The macro also checks for the file
 *   handle so it doesn't do the snprintf if the log is not enabled.
 * Parameters: none
 * External References:
 *   LOGfh: Log file stream handle opened below.
 *   logMsg[]: The string to write to the log. We also add the date/time to each message.
 * Returned values: none
 * Errors detected: none
 */
{
  // Punt if no log file.
  if (LOGfh == NULL) return;

  // Write the entry with the date and time.
  fprintf(LOGfh, "%s: %s\n", date(), logMsg);
  fflush(LOGfh);
}



void initLog(void)
/*     If a name is given for the log file (usually a directory), then create a file name, open
 *   it for output.
 * Parameters: none
 * External References:
 *   MAXPATHLEN: OS dependent, constant.
 *   LOGFileName: changed from the command line, usually a directory.
 *   LOGfh: The file handle for logging if the file is opened.
 * Returned values: none
 * Errors detected: Couldn't open file, results in a halt.
 */ 
{
  char fname[MAXPATHLEN];

  // Open the log file.
  if (LOGFileName != NULL)
  {
    snprintf(fname, MAXPATHLEN, "%s%s.mqtt", LOGFileName, date());
    if ((LOGfh = fopen(fname, "wb")) == NULL)
    {
      fprintf(stderr, "Couldn't open log file %s\n", fname);
      exit(-1);
    }
  }
  else LOGfh = NULL;
}

// EOF
