/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: mail.c
 * Author: Jed Marti
 * Description: Abbreviations for From/To field. Search a list and return the expanded or
 *   original.
 * Created: Mon Dec  2 10:57:37 2024
 * Revision History:
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>

#include "mqtt.h"

// Output noise level.
extern uint8_t Noise;

// A structure for storing lists of abbreviations.
typedef struct {
  void *next;         // A linked list.
  char *abbrev;       // The abbreviation.
  char *subst;        // What it expands to.
} abbreviation;


// This is a list read from a file at startup.
abbreviation *A = NULL;


static void newAbbreviation(char *abbrev, char *subst)
/*     Create a new abbreviation structure and add it to the list. This is not a Generic so it
 *   can't be displayed or GC'd.
 * Parameters:
 *   abbrev: The string from the file with the abbreviation name.
 *   subst: What to substitute. An e-mail address or AREDN IPV4 number.
 * External References:
 *   A: the list of abbreviations. This new one added to the front of the list.
 * Returned values: none.
 * Errors detected: out of space halt.
 */
{
  abbreviation *a;
  if ((a = (abbreviation*) malloc(sizeof(abbreviation))) == NULL)
  {
    fprintf(stderr, "%s:%s:%d Out of space for abbreviation\n", __FILE__, __func__, __LINE__);
    exit(-1);
  }

  // Fill in and add to the front of the abbreviation list.
  a->next = NULL,
    a->abbrev = abbrev,
    a->subst = subst;
  a->next = A;
  A = a;
}



void initAbbreviations(char *fname)
/*     Initialize the abbreviation list from a file somewhat like /etc/hosts.
 * Parameters:
 *   fname: File name to read. If not found, signal error but let program continue.
 * External References:
 *   Noise: How much notse to generate.
 *   MAXPATHLEN: Largest reasonable file size.
 * Returned values: none
 * Errors detected: No file gives error message but continues.
 */
{
  char buf[MAXPATHLEN];
  FILE *fh;
  char *ab;
  char *su;

  if ((fh = fopen(fname, "rb")) == NULL)
  {
    fprintf(stderr, "%s:%s:%d Can't open %s for output\n", __FILE__, __func__, __LINE__, fname);
    return;
  }
  if (Noise > 1)
 printf("%s:%s:%d Reading Abbrevation file %s\n", __FILE__, __func__, __LINE__, fname);
  
 loop:
  // End of file means quit.
  if (feof(fh) || (fgets(buf, MAXPATHLEN, fh) == NULL))
  {
    fclose(fh);
    return;
  }

  // Comment, but no blank lines.
  if (buf[0] == '#') goto loop;

  // Separated by tabs. Note that the next string has an actual tab in it so don't remove it.
  ab = strdup(strtok(buf, "	"));
  su = strdup(strtok(NULL, " \n"));
  if (Noise > 2)  printf("  %s:%s:%d %s becomes %s\n", __FILE__, __func__, __LINE__, ab, su);
  newAbbreviation(ab, su);
  goto loop;
}



char *expandAbbreviation(char *abb)
/*     We have a from or to address. See what it should really be.
 * Parameters:
 *   abb: from an NQTT message or net service.
 * External References:
 *   A: The list of abbreviations.
 * Returned values: The substitute or the original. This is not a copy so if you put it into
 *   a generic string you'll corrupt the system.
 * Errors detected: none
 */
{
  abbreviation *a;
  for (a = A; a != NULL; a = (abbreviation*) (a->next))
    if (strcasecmp(a->abbrev, abb) == 0) return a->subst;
  return abb;
}

    
// EOF
