/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: mprotos.c
 * Author: Jed Marti
 * Description: protobuf definitions from buf.build.meshtastic/protobufs. These are hand compiled
 *   for this system - not all are used. These are the contents of the MQTT PUBLISH message.
 * Created: Sat Dec  7 10:17:15 2024
 * Revision History:
 *   Sun Dec 29 15:42:53 2024 - Code review.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>

#include "mqtt.h"
#include "protobuf.h"

// Local definition for message table.
typedef void (*pparse)(int32_t portnum, Generic *p);

// Noise setting.
extern uint8_t Noise;

// Log files, see log.c, used by LOG macro even if they don't show up here.
extern char logMsg[];
extern FILE *LOGfh;


// A table of functions based on the portnum part of a Data packet and how to parse the message
// contained therein. Up to 511 are allowed but only a few get sent to MQTT. 
pparse FMTS[MAX_APP + 1];



void unknownProtoBuf(int32_t portnum, Generic *p)
/*     If we get a PUBLISH MQTT message, its portnum field says how to decode it. Of the 511
 *   allowed, only a few show up on MQTT so the other 500 or so end up here.
 * Parameters:
 *   portnum: Integer from the Data packet portnum field.
 *   p: The PUBLISH packet a-list (think JSON).
 * External References: none
 * Returned values: does not return. Later we should probably write to a log file and check
 *   for it filling up.
 * Errors detected: Log and continue.
 */
{
  fprintf(stderr, "%s:%s:%d Unknown protoBuf message descriptor (portnum) 0x%02X\n",
          __FILE__, __func__, __LINE__, portnum);
  LOG("Unknown protoBuf message descriptor 0x%02X\n", portnum);
}


void textMessage(int32_t portnum, Generic *p)
/*     The message gets passed to sendmail - it's just text. Send mail will decide where to
 *   send it by checking the current abbreviation list and format. The block needs both MQTT
 *   and protoBuf data.
 * Parameters:
 *   portnum: Integer from the Data packet portnum field.
 *   p: The PUBLISH packet a-list (think JSON).
 * External References: none
 * Returned values: none
 * Errors detected: none
 */
{
  sendMail(p);
}


void ignore(int32_t portnum, Generic *p)
/*     Some messages just don't mean much so they get ignored rather than signaling an error.
 * Parameters:
 *   portnum: Integer from the Data packet portnum field.
 *   p: The PUBLISH packet a-list (think JSON).
 * External References:
 *   Noise: If > 1 then display that we're ignoring the message.
 * Returned values: none
 * Errors detected: none
 */
{
  if (Noise > 1)
    printf("%s:%s:%d ignoring message type %d\n", __FILE__, __func__, __LINE__, portnum);
}


// The Container for various messages that wander around the system.
protoBuf ServiceEnvelope[] = {
    {.gtype=SGROUP, .index=1, .name="packet", .sub=MeshPacketId},
    {.gtype=LEN, .ptype=proto_len, .index=2, .name="channel_id", .sub=noSub},
    {.gtype=LEN, .ptype=proto_len, .index=3, .name="gateway_id", .sub=noSub},
    {.index=-1}   };
protoBlock protoServiceEnvelope = {
  .name="ServiceEnvelope",
  .p = ServiceEnvelope };


// Container for a packet sent around the net.
protoBuf MeshPacket[] =  {
  {.name="from", .index=1, .gtype=I32, .ptype=proto_fixed32, .sub=noSub},
  {.name="to", .index=2, .gtype=I32, .ptype=proto_fixed32, .sub=noSub},
  {.name="channel", .index=3, .gtype=I32, .ptype=proto_uint32, .sub=noSub},
  {.name="Data", .index=4, .gtype=SGROUP, .sub=DataId},
  {.name="id", .index=6, .gtype=I32, .ptype=proto_fixed32, .sub=noSub},
  {.name="rx_time", .index=7, .gtype=I32, .ptype=proto_fixed32, .sub=noSub},
  {.name="rx_snr", .index=8, .gtype=I32, .ptype=proto_float, .sub=noSub},
  {.name="hop_limit", .index=9, .gtype=I32, .ptype=proto_uint32, .sub=noSub},
  {.name="want_ack", .index=10, .gtype=VARINT, .ptype=proto_bool, .sub=noSub},
  {.name="priority", .index=11, .gtype=VARINT, .ptype=proto_enum, .sub=noSub},
  {.name="rx_ssi", .index=12, .gtype=I32, .ptype=proto_int32, .sub=noSub},
  {.name="delayed", .index=13, .gtype=VARINT, .ptype=proto_enum, .sub=noSub},
  {.name="via_mqtt", .index=14, .gtype=VARINT, .ptype=proto_bool, .sub=noSub},
  {.name="hop_start", .index=15, .gtype=VARINT, .ptype=proto_uint32, .sub=noSub},
  {.name="public_key", .index=16, .gtype=ERROR, .sub=noSub},
  {.name="pki_encrypted", .index=17, .gtype=VARINT, .ptype=proto_bool, .sub=noSub},
  {.index=-1} };
protoBlock protoMeshPacket = {.name="MeshPacket", .p=MeshPacket};


// The data packet. The real stuff is determined by the portnum field and the payload
// is specific to the type of message.
protoBuf Data[] = {
  {.name="portnum", .index=1, .gtype=VARINT, .ptype=proto_enum, .sub=noSub},
  {.name="payload", .index=2, .gtype=LEN, .ptype=proto_len, .sub=noSub },
  {.name="want_response", .index=3, .gtype=VARINT, .ptype=proto_bool, .sub=noSub},
  {.name="dest", .index=4, .gtype=I32, .ptype=proto_fixed32, .sub=noSub},
  {.name="source", .index=5, .gtype=I32, .ptype=proto_fixed32, .sub=noSub},
  {.name="request_id", .index=6, .gtype=I32, .ptype=proto_fixed32, .sub=noSub},
  {.name="reply_id", .index=7, .gtype=I32, .ptype=proto_fixed32, .sub=noSub},
  {.name="emoji", .index=8, .gtype=I32, .ptype=proto_fixed32, .sub=noSub},
  {.name="bitfield", .index=9, .gtype=VARINT, .ptype=proto_uint32, .sub=noSub},
  {.index=-1}};
protoBlock protoData = {.name="Data", .p=Data};

// Never seen it but here for fun.
protoBuf BluetoothConnectionStatus[] = {
  {.name="pin", .index=1, .gtype=VARINT, .ptype=proto_uint32, .sub=noSub},
  {.name="rssi", .index=2, .gtype=VARINT, .ptype=proto_int32, .sub=noSub},
  {.name="is_connected", .index=3, .gtype=VARINT, .ptype=proto_bool, .sub=noSub},
  {.index=-1}};
protoBlock protoBluetoothConnectionStations =
  { .name="bluetoothConnectionStatus", .p=BluetoothConnectionStatus};

// Information about a node intended to be reported unencrypted to a map using MQTT.
protoBuf MapReport[] = {
  {.name="long_name", .index=1, .gtype=LEN, .ptype=proto_len, .sub=noSub},
  {.name="short_name", .index=2, .gtype=LEN, .ptype=proto_len, .sub=noSub},
  {.name="role", .index=3, .gtype=VARINT, .ptype=proto_enum, .sub=noSub},
  {.name="hw_model", .index=4, .gtype=VARINT, .ptype=proto_enum, .sub=noSub},
  {.name="firmware_version", .index=5, .gtype=LEN, .ptype=proto_len, .sub=noSub},
  {.name="region", .index=6, .gtype=VARINT, .ptype=proto_enum, .sub=noSub},
  {.name="modem_preset", .index=7, .gtype=VARINT, .ptype=proto_enum,  .sub=noSub},
  {.name="has_default_channel", .index=8, .gtype=VARINT, .ptype=proto_bool, .sub=noSub},
  {.name="latitude_i", .index=9, .gtype=I32, .ptype=proto_sfixed32, .sub=noSub},
  {.name="longitude_i", .index=10, .gtype=I32, .ptype=proto_sfixed32, .sub=noSub},
  {.name="altitude", .index=11, .gtype=VARINT, .ptype=proto_int32, .sub=noSub},
  {.name="position_precision", .index=12, .gtype=VARINT, .ptype=proto_uint32, .sub=noSub},
  {.name="num_online_local_nodes", .index=13, .gtype=VARINT, .ptype=proto_uint32, .sub=noSub}};
protoBlock protoMapReport = {.name="MapReport", .p=MapReport};



void mapReportAppProtoBuf(int32_t portnum, Generic *p)
/*     Names of objects and potential positions. Currently a dummy.
 * Parameters:
 *   portnum: Integer from the Data packet portnum field.
 *   p: The PUBLISH packet a-list (think JSON).
 * External References:
 *   Noise: If > 1 then display that we're ignoring the message.
 * Returned values: none
 * Errors detected: none
 */
{
  printf("Map report App\n");
}
      
// A GPS position. This appears a lot and helps us with our position and others.
protoBuf Position[] = {
  {.name="latitude_i", .index=1, .gtype=I32, .ptype=proto_sfixed32, .sub=noSub},
  {.name="longitude_i", .index=2, .gtype=I32, .ptype=proto_sfixed32, .sub=noSub},
  {.name="altitude", .index=3, .gtype=VARINT, .ptype=proto_int32, .sub=noSub},
  {.name="time", .index=4, .gtype=I32, .ptype=proto_fixed32, .sub=noSub},
  {.name="location_source", .index=5, .gtype=VARINT, .ptype=proto_enum, .sub=noSub},
  {.name="altitude_source", .index=6, .gtype=VARINT, .ptype=proto_enum, .sub=noSub},
  {.name="timestamp", .index=7, .gtype=I32, .ptype=proto_fixed32, .sub=noSub},
  {.name="timestamp_millis_adjust", .index=8, .gtype=I32, .ptype=proto_sfixed32, .sub=noSub},
  {.name="altitude_hae", .index=9, .gtype=VARINT, .ptype=proto_sint32, .sub=noSub},
  {.name="altitude_geoidal_separation", .index=10, .gtype=VARINT, .ptype=proto_sint32, .sub=noSub},
  {.name="PDOP", .index=11, .gtype=VARINT, .ptype=proto_uint32, .sub=noSub},
  {.name="HDOP", .index=12, .gtype=VARINT, .ptype=proto_uint32, .sub=noSub},
  {.name="VDOP", .index=13, .gtype=VARINT, .ptype=proto_uint32, .sub=noSub},
  {.name="gps_accuracy", .index=14, .gtype=VARINT, .ptype=proto_uint32, .sub=noSub},
  {.name="ground_speed", .index=15, .gtype=VARINT, .ptype=proto_uint32, .sub=noSub},
  {.name="ground_track", .index=16, .gtype=VARINT, .ptype=proto_uint32, .sub=noSub},
  {.name="fix_quality", .index=17, .gtype=VARINT, .ptype=proto_uint32, .sub=noSub},
  {.name="fix_type", .index=18, .gtype=VARINT, .ptype=proto_uint32, .sub=noSub},
  {.name="sats_in_view", .index=19, .gtype=VARINT, .ptype=proto_uint32, .sub=noSub},
  {.name="sensor_id", .index=20, .gtype=VARINT, .ptype=proto_uint32, .sub=noSub},
  {.name="next_update", .index=21, .gtype=VARINT, .ptype=proto_uint32, .sub=noSub},
  {.name="seq_number", .index=22, .gtype=VARINT, .ptype=proto_uint32, .sub=noSub},
  {.name="precision_bits", .index=23, .gtype=VARINT, .ptype=proto_uint32, .sub=noSub},
  {.index=-1}};
protoBlock protoPosition = {.name="Position", .p=Position};

// A shorter version of Position.
protoBuf PositionLite[] = {
  {.name="latitude_i", .index=1, .gtype=I32, .ptype=proto_sfixed32, .sub=noSub},
  {.name="lontitude_i", .index=2, .gtype=I32, .ptype=proto_sfixed32, .sub=noSub},
  {.name="altitude", .index=3, .gtype=VARINT, .ptype=proto_int32, .sub=noSub},
  {.name="time", .index=4, .gtype=I32, .ptype=proto_fixed32, .sub=noSub},
  {.index=-1}};
protoBlock protoPositionLite = {.name="PositionLite", .p=PositionLite};

static void positionAppProtoBuf(int32_t portnum, Generic *p)
/*     Decide on lite or full.
 */
{
  position(p);
}

// Broadcast when a newly powered mesh node wants to find a node num it can use
// Sent from the phone over bluetooth to set the user id for the owner of this node.
// Also sent from nodes to each other when a new node signs on (so all clients can have this info)
// The algorithm is as follows:
// when a node starts up, it broadcasts their user and the normal flow is for all
// other nodes to reply with their User as well (so the new node can build its nodedb)
// If a node ever receives a User (not just the first broadcast) message where
// the sender node number equals our node number, that indicates a collision has
// occurred and the following steps should happen:
// If the receiving node (that was already in the mesh)'s macaddr is LOWER than the
// new User who just tried to sign in: it gets to keep its nodenum.
// We send a broadcast message of OUR User (we use a broadcast so that the other node can
// receive our message, considering we have the same id - it also serves to let
// observers correct their nodedb) - this case is rare so it should be okay.
// If any node receives a User where the macaddr is GTE than their local macaddr,
// they have been vetoed and should pick a new random nodenum (filtering against
// whatever it knows about the nodedb) and rebroadcast their User.
// A few nodenums are reserved and will never be requested:
// 0xff - broadcast
// 0 through 3 - for future use
protoBuf User[] = {
  {.name="id", .index=1, .gtype=LEN, .ptype=proto_len, .sub=noSub},
  {.name="long_name", .index=2, .gtype=LEN, .ptype=proto_len, .sub=noSub},
  {.name="short_name", .index=3, .gtype=LEN, .ptype=proto_len, .sub=noSub},
  {.name="macaddr", .index=4, .gtype=LEN, .ptype=proto_len, .sub=noSub},
  {.name="hw_model", .index=5, .gtype=VARINT, .ptype=proto_uint32, .sub=noSub},
  {.name="is_licensed", .index=6, .gtype=VARINT, .ptype=proto_bool, .sub=noSub},
  {.name="role", .index=7, .gtype=VARINT, .ptype=proto_uint32, .sub=noSub},
  {.name="public_key", .index=8, .gtype=LEN, .ptype=proto_len, .sub=noSub} };
protoBlock protoUser = {.name="User", .p=User};



////////////////////////////////////////////////////////////////////////////////////////////////////
// Just for testing, does not appear from Meshtastic.
protoBuf testPacket[] = {
  {.name="part1", .index=1, .gtype=SGROUP, .sub=testSubId},
  {.name="end", .index=2, .gtype=VARINT, .ptype=proto_int32, .sub=noSub},
  {.index=-1} };
protoBlock protoTestPacket = {.name="testPacket", .p=testPacket};

protoBuf testSub[] = {
  {.name="subInt", .index=1, .gtype=VARINT, .ptype=proto_int32, .sub=noSub},
  {.index=-1}
};
protoBlock protoTestSub = {.name="testSub", .p=testSub};
////////////////////////////////////////////////////////////////////////////////////////////////////


// A table of blocks.
protoBlock *subBlocks[] =
  { NULL, &protoServiceEnvelope, &protoMeshPacket, &protoData, &protoPositionLite,
    &protoUser, &protoTestSub, &protoMapReport };


// A table of potential packets, NULL at end indicates end.
protoBlock *potentials[] = {&protoPositionLite, &protoUser, &protoPosition, &protoMapReport,  NULL};


protoBlock *sub2Block(uint8_t sub)
/*     Get the ProtoBlock by the index. Our table says we need a subblock, locate it so we can
 *   recursively get the data nested as deep as you want.
 * Parameters:
 *   sub: index into subBlocks.
 * External References:
 *   subBlocks[]: A vector of blocks (see above).
 * Returned values: The control block for the protoBuf.
 */
{
  if (sub > (sizeof(subBlocks) / sizeof(protoBlock*)))
  {
    fprintf(stderr, "%s:%s:%d Sub %d out of subBlocks range\n", __FILE__, __func__, __LINE__, sub);
    exit(-1);
  }
  return subBlocks[sub];
}



void farmOutPayload(Generic *g)
/*     We have a packet from the ServiceEnvelope stored as a JSON like list. Farm out what
 *   we're supposed to do with the data.
 * Parameters:
 *   g: The ServiceEnvelope data structure as an association list of (field-name . value)
 * External References:
 *   Noise: How much to babble.
 *   TESTVAL, NULLCHK: macros for error testing on Generics - see mqtt.h
 * Returned values: none
 * Errors Detected: No data field, missing data elements, unknown application.
 */
{
  Generic *pkt, *data, *payload, *portnum;
  
  // "packet" comes from the Service Envelope.

  pkt = assoc("packet", g);
  if (Noise > 4)
  {
    printf("%s:%s:%d packet =", __FILE__, __func__, __LINE__);
    display(stdout, pkt);
    printf("\n");
  }

  // The data component contains the message descriptor and the message descriptor. We've
  // already detected that payload contains the text message. An empty messages causes an
  // early termination.
  data = assoc("data", pkt);
  if (data == NULL)
  {
    LOG("Message has no data field");
    printf("%s:%s:%d Message has no data field\n", __FILE__, __func__, __LINE__);
    return;
  }
  NULLCHK(data, "No data payload");
  portnum = assoc("portnum", data);
  NULLCHK(portnum, "No message type");
  payload = assoc("payload", data);
  NULLCHK(payload, "Empty text message");
  if (!__isInt(portnum))
  {
    display(stderr, data);
    fprintf(stderr, "%s:%s:%d Message type not found\n", __FILE__, __func__, __LINE__);
    exit(-1);
  }

  // Farm it out.
  TESTVAL(__int(portnum) <= MAX_APP, "Illegal application message");
  (FMTS[__int(portnum)])(__int(portnum), g);

  // At this point g should be GC'd as we're done with it though the caller probably ought to
  // do this. See publish.c for where this is done.
}


void initProtoBufs(void)
/*     Initialize the table of what to do with each portnum. Up to 511 are allowed only a few
 *   of which are used so we make the others illegal.
 * Parameters: none
 * External References:
 *   FMTS[]: indicate what function to call to decode the message.
 * Returned values: none
 * Errors detected: none
 */
{
  int32_t i;

  // Default message is unknown type.
  for (i = 0; i <= MAX_APP; ++i) FMTS[i] = unknownProtoBuf;

  FMTS[TEXT_MESSAGE_APP] = textMessage;
  FMTS[POSITION_APP] = positionAppProtoBuf;
  FMTS[NODEINFO_APP] = userMessage;
  FMTS[REPLY_APP] = ignore;
}


// EOF
