/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: mqtt.h
 * Author: Jed Marti
 * Description: Prototypes etc. for light MQTT server.
 * Created: Tue Dec  3 08:09:47 2024
 * Revision History:
 *   Mon Dec 30 12:12:02 2024 - Code review
 */

#ifndef mqtt_h
#define mqtt_h

#define VERSION 2
#define RELEASE 6
#define PATCH   0


/////////////////////////// MAXIMUM SIZES AND STARTUP DEFAULTS /////////////////////////////////////

// Input buffer size.
#define NBUFFERSIZE 2048

// Maximum message size (overkill)
#define MAXMSG 1024

// For file names. File names should not be essays with blanks. 
#ifndef MAXPATHLEN
  #define MAXPATHLEN 2048
#endif

// Maximum name size for various parts of messages.
#define MAXNAMESIZE 128

// MQTT default port number - for non encrypted messages.
#define DEFAULT_PORTNUM 1883

// AREDN mail server portnumber
#define DEFAULT_AREDN  3976

// Startup add time for watchdog. Extra time in seconds.
#define KEEPALIVESTARTUP 30

// Default SMS text message directory.
#define SMSDIR "/tmp/"

// Log file name.
#define LOGNAME ".log/"

// Mail channel.
#define ARESCHANNEL "ARESMAIL"

// Packet tracing directory. 
#define TRACEDIR "/tmp/"

////////////////////////////// MQTT STRUCTURES AND ENUMERATIONS ////////////////////////////////////

// MQTT message descriptors in the top half of the first byte.
enum {
      MQTT_UNKNOWN0    = 0x00,
      MQTT_CONNECT     = 0x01,
      MQTT_CONNACK     = 0x02,
      MQTT_PUBLISH     = 0x03,
      MQTT_PUBACK      = 0x04,
      MQTT_PUBREC      = 0x05,
      MQTT_PUBREL      = 0x06,
      MQTT_PUBCOMP     = 0x07,
      MQTT_SUBSCRIBE   = 0x08,
      MQTT_SUBACK      = 0x09,
      MQTT_UNSUBSCRIBE = 0x0A,
      MQTT_UNSUBACK    = 0x0B,
      MQTT_PINGREQ     = 0x0C,
      MQTT_PINGRESP    = 0x0D,
      MQTT_DISCONNECT  = 0x0E,
      MQTT_UNKNOWNF    = 0x0F};


// For connection status,
enum {
      downNoConnection,
      downUpConnected,
      downDownPending,
      downConnected
};


// Storing topic strings as a list.
typedef struct {
  void *next;          // Stored as a list.
  char *topic;         // The ASCII topic string.
} Topic;



//////////////////////////////// LOG AND ERROR MACROS //////////////////////////////////////////////

// Logging stuff. Looks like printf.
#define LOG(format, ...) \
{        \
  if (LOGfh)          \
  {                   \
    snprintf(logMsg, MAXPATHLEN*2, format __VA_OPT__(,) __VA_ARGS__);        \
    iLog();          \
  }                  \
}


// Some error message hacks.
#define NULLCHK(x, var)                                \
  if ((x) == NULL) \
  {                            \
     fprintf(stderr, "%s:%s:%d %s is NULL", __FILE__, __func__, __LINE__, var); \
     exit(-1);           \
  }
#define TESTVAL(tst, msg) \
  if (!(tst))                                        \
  { \
    fprintf(stderr, "%s:%s:%d %s", __FILE__, __func__, __LINE__, msg);  \
    exit(-1); \
  }

//////////////////////////////////// MODULE PROTOTYPES /////////////////////////////////////////////


// Parse an MQTT packet prior to protoBuf parsing, farm out work. See packet.c
void parseMQTTPacket(void);

// Copy a message to the simple AREDN server. Mail is to an IPV4 address.See arednsend.c
void arednSend(char *rmt, int32_t portnum, char *fname);

// Parse the command line arguments and help. See cmdline.c
void cmdline(int32_t argc, char *argv[]);

// An MQTT CONNECT message has arrived. See connect.c
void mqttConnect(uint8_t pt, uint8_t *pkt);

// Returns date/time as a string. See date.c.
char *date(void);

// MQTT sender has initiated a DISCONNECT message - clean up. See disconnect.c
void mqttDisconnect(uint8_t pt, uint8_t *pkt);

// See log.c
void iLog(void);
void initLog(void);

// See listen.c
void netListen(int32_t portnum);
void netAccept(void);

// See mail.c
void initAbbreviations(char *fname);
char *expandAbbreviation(char *abb);

// See meshtastic.o
void parseMeshtastic(void);

// See net.c
uint8_t ngetch(void);
uint8_t pgetch(uint8_t *pkt);
uint16_t pget16(uint8_t *pkt);
void nputch(uint8_t v);
void nflush(void);

// See netconnect.c
int32_t netConnect(char *pname, int32_t pno, uint32_t tout);

// Write a binary packet to a file. See packettr.c.
void packetTrace(uint8_t ptype, uint8_t *pkt, int32_t pl);

// See pingreq.c
void mqttPingReq(uint8_t pt, uint8_t *pkt);
  
// See publish.c
void mqttPublish(uint8_t pt, uint8_t *pkt, int32_t rl);
void mqtt2Mesh(uint8_t *msg, int32_t msize);

// See pubcomp.c
void mqttPubComp(uint8_t pt, uint8_t *pkt);

// See ssmtp.c
void checkGateways(void);

// See subscribe.c
char *whichIs(const char *ba);
void clearTopics(void);
void mqttSubscribe(uint8_t pt, uint8_t *pkt);

// See textmsg.c
void text2protoBuf(char *msg);
void publishMessage(char *msg);

#endif


// EOF
