/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: net.c
 * Author: Jed Marti
 * Description: Basic network functionality to make TCP/IP look like a stream. Note that MQTT
 *    packets read direct, but that MQTT publish packets use a different strategy.
 * Created: Mon Dec  2 06:23:17 2024
 * Revision History:
 *   Sun Dec 29 16:22:58 2024 - Code review.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>


#include "mqtt.h"

// The  input stream buffer, index and read size. We read a message here to avoid TCP problems
// and then treat it like a stream.
uint8_t B[NBUFFERSIZE];
int32_t BINDEX = -1;
int32_t BSIZE = 0;

// The output stream buffer. Used for sending some messages to MQTT (which it seems to ignore).
uint8_t O[NBUFFERSIZE];
int32_t OINDEX = 0;


// File handle set by listen.c. If -1, not assigned.
extern int32_t FD;

// Noise level 0-shut up, > 0 more and more babble.
extern uint8_t Noise;

// For simulating packet get - used for testing.
int32_t packetIndex;


uint8_t pgetch(uint8_t *pkt)
/*     Get the next byte from the current TCP/IP buffer.
 * Parameters:
 *   pkt: Address of the current buffer (probably B but this is old code).
 * External References:
 *   packetIndex: Index into this buffer.
 * Returned values: next byte.
 * Errors detected: none, should probably check for running off end.
 */
{
  return pkt[packetIndex++];
}


uint16_t pget16(uint8_t *pkt)
/*     Get 16 bit unsigned int from 2 bytes. big-endian. This is used by MQTT parsing only,
 *   not protoBufs.
 * Parameters:
 *   pkt: packet buffer.
 * External References: none
 * Returned values: an unsigned 16 bit integer. Usually a string length.
 * Errors detected: none.
 */
{
  uint16_t r;
  r = ((uint16_t) pgetch(pkt)) << 8;
  r |= (uint16_t) pgetch(pkt);
  return r;
}



uint8_t ngetch(void)
/*     Return the next character from the input buffer. If the buffer is empty read a new one.
 * Parameters: none
 * External References:
 *   FD: file handle to read from.
 *   B[]: An input buffer.
 *   BINDEX: Index into the buffer. If -1, then buffer is empty.
 *   BSIZE: How many bytes in the buffer.
 * Returned values: The next byte or wait for one.
 * Errors Detected: FD not set yet. 
 */
{
  ssize_t n;

  // Shouldn't happen.
  if (FD < 0)
  {
    fprintf(stderr, "Not connected to any remote\n");
    exit(-1);
  }
  
  // Buffer empty, read a new one.
 loop:
  if (BINDEX == -1 || BINDEX >= BSIZE)
  {
    // Read, but if nothing, wait a while and try again.
    n = read(FD, B, NBUFFERSIZE);
    if (n == 0)
    {
      usleep(1000);
      goto loop;
    }

    // Here the remote send something really huge we can't deal with.
    if (n == NBUFFERSIZE)
    {
      fprintf(stderr, "%s:%s:%d Buffer overflow\n", __FILE__, __func__, __LINE__);
      exit(-1);
    }

    // Return the first character.
    if (Noise > 6) printf("%s:%s:%d Read %ld\n", __FILE__, __func__, __LINE__, n);
    BSIZE = (int32_t) n;
    BINDEX = 1;
    return B[0];
  }

  // Otherwise we still have characters to return.
  return B[BINDEX++];
}


////////////////////////////////////////////////////////////////////////////////////////////////////
//                                 Us -> Meshtastic


void nputch(uint8_t v)
/*      Save a byte in the buffer.
 * Parameters:
 *   v: byte to save
 * External References:
 *   OINDEX: index into output buffer.
 *   NBUFFSERSIZE: constant - maximum buffer size.
 *   O[]: The buffer.
 * Returned values: none
 * Errors detected: buffer overflow - causes halt.
 */
{
  if (OINDEX >= NBUFFERSIZE)
  {
    fprintf(stderr, "%s:%s:%d Network output buffer overflow\n", __FILE__, __func__, __LINE__);
    exit(-1);
  }
  O[OINDEX++] = v;
}



void nflush(void)
/*     Flush the network output buffer by sending it to the meshtastic device: WiFi/Ethernet.
 * Parameters: none
 * External References:
 *   OINDEX: index into output buffer.
 *   O[]: The buffer.
 *   Noise: if > 4 then show write results.
 * Returned values: none
 * Errors detected: write fails, causes halt.
 */
{
  if (OINDEX > 0)
  {
    if (write(FD, O, OINDEX) < 0)
    {
      fprintf(stderr, "%s:%s:%d Network write fails because %s\n", __FILE__, __func__, __LINE__,
              strerror(errno));
      exit(-1);
    }
    if (Noise > 4)
      printf("%s:%s:%d Write %d bytes to %d\n", __FILE__, __func__, __LINE__, OINDEX, FD);
    OINDEX = 0;
  }
}
    
      
  
// EOF
