/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: netconnect.c
 * Author: Jed Marti
 * Description: Connect to a remote, return port number. This is used to connect to a remote
 *   simple mail server when the target mail is an IPV4 address. This is not SMTP.
 * Created: Tue Dec  3 16:56:43 2024
 * Revision History:
 *   Sun Dec 29 17:13:40 2024 - Code review.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <setjmp.h>
#include <signal.h>
#include <netdb.h>
#include <unistd.h>

#include "mqtt.h"

// Here's a jump address for when the connect to remote fails.
jmp_buf connect_err;



void connectcatcher(int32_t v)
/*     Gets called when netConnect times out. Jump back to the caller that detects the error.
 * Parameters:
 *   v: Value to return from setjmp.
 * External References:
 *   connect_err: For Linux saving stack and so on.
 * Returned values: none, does not return.
 * Errors detected: none
 */
{
  longjmp(connect_err, v);
}



int32_t netConnect(char *pname, int32_t pno, uint32_t tout)
/*     Connect to a remote and return our handle.
 * Parameters:
 *   pname: An IPV4 string or DNS name.
 *   pno: Port number on remote. Should be > 1024 to avoid problems with root user.
 *   tout: Timeout wait for remote to answer in seconds.
 * External References:
 *   jmp_buf: Where to go on alarm time out.
 *   errno: Error number from system.
 * Returned value: File handle for TCP I/O or -1 for fail.
 * Errors detected: opening fails, time out.
 */
{
  struct  hostent *host;
  int32_t sock = -1;
  struct  sockaddr_in my_addr;  

  // Get the IP number to connect to.
  host = gethostbyname(pname);
  if (host == NULL)
  {
    fprintf(stderr, "%s:%s: %d %s invalid or unknown host name or number\n", __FILE__,
            __func__, __LINE__, pname);
    return -1;
  }
     
  // Timeout.
  signal(SIGALRM, connectcatcher);
  alarm(tout);
  if (setjmp(connect_err) == 0)
  {
    // Get the socket.
    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0)
    {
      fprintf(stderr, "%s:%s:%d open fails because %s\n", __FILE__, __func__, __LINE__,
              strerror(errno));
      exit(-1);
    }

    // Build the address for the connection call.
    memset(&my_addr, 0, sizeof(my_addr));
    my_addr.sin_family = AF_INET;
    memcpy((char *)&my_addr.sin_addr, host -> h_addr_list[0], host -> h_length);
    my_addr.sin_port = htons((short) pno);

    if (connect(sock, (struct sockaddr *) &my_addr, sizeof (my_addr)) < 0)
    {
      close(sock);
      alarm(0);
      signal(SIGALRM, SIG_IGN);
      fprintf(stderr, "%s:%s:%d fails because %s]n", __FILE__, __func__, __LINE__, strerror(errno));
      return  -1;
    }
  }

  // If we get here, we timed out.
  else
  {
    fprintf(stderr, "%s:%s:%d Connection to %s:%d timed out\n", __FILE__, __func__, __LINE__,
            pname, pno);
    sock = -1;
  }
  
  // Turn off the alarm and signal.
  alarm(0);
  signal(SIGALRM, SIG_IGN);

  return sock;
}


// EOF
