/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: packet.c
 * Author: Jed Marti
 * Description: Farm out packets to the various routines. Works only for MQTT 3.1.1
 * Created: Mon Dec  2 06:07:45 2024
 * Revision History:
 *   Thu Dec  5 10:04:03 2024 - Add packet tracing.
 *   Sun Dec 29 17:22:16 2024 - Code review.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <unistd.h>

#include "mqtt.h"

// Noise level.
extern uint8_t Noise;

// Packet tracing.
extern uint8_t packetTr;

// Packet index into the input buffer.
extern int32_t packetIndex;

// Log files, see log.c, used by LOG macro.
extern char logMsg[];
extern FILE *LOGfh;

// For debugging. These correspond to the MQTT description 3.1.1. The names correspond to the
// top 4 bits of the MQTT packet descriptor. This is receive only. Some of these are responses to
// packets sent back to the Meshtastic device.
char *mqttTypes[] = { "0x0",
                    "CONNECT",
                    "CONNACK",
                    "PUBLISH",
                    "PUBACK",
                    "PUBREC",
                    "PUBREL",
                    "PUBCOMP",
                    "SUBSCRIBE",
                    "SUBACK",
                    "UNSUBSCRIBE",
                    "UNSUBACK",
                    "PINGREQ",
                    "PINGRESP",
                    "DISCONNECT",
                    "0xF"};


void parseMQTTPacket(void)
/*     Read an MQTT 3.1.1 packet. We know the type from the first few bytes so farm the message
 *   out to each reader/responder.
 * Parameters: none
 * External References:
 *   packetTr: Trace packets through file dump.
 *   Noise: How much noise to generate.
 *   packetIndex: TCP input buffer index.
 * Returned values: none
 * Errors detected: Packet too large (has VARINT size). Packets are supposed to be < 255.
 *   An unknown packet type is flagged but we skip over it.
 */
{
  uint8_t pt;
  uint8_t byt;
  int32_t rl;
  uint8_t pkt[MAXMSG];
  int32_t i;

  // Packet descriptor.
  pt = ngetch();

  // 1-2 byte length. Standard says 4 but we're talking to something small without 256 MB of
  // storage.
  //  0-127 is a single byte length.
  byt = ngetch();

  // High bit set means 7 more bits follow.
  if (byt & 0x80)
  {
    rl = byt & 0x7F;
    // A 3rd byte will cause a failure.
    byt = ngetch();
    if (byt & 0x80)
    {
      fprintf(stderr, "%s:%s:%d 3 or more byte packet length type %s\n", __FILE__, __func__,
              __LINE__, mqttTypes[pt]);
      exit(-1);
    }
    // This will allow up to 16k byte packets.
    rl += ((int32_t) byt) * 128;
  }

  // Just a single byte length.
  else rl = byt;

  // Collect the packet.
  for (i = 0; i < rl; ++i) pkt[i] = ngetch();

  // Trace code.
  if (Noise > 2) printf("%s:%s:%d Packet type %s, size %d\n", __FILE__, __func__, __LINE__,
                        mqttTypes[pt >> 4], rl);
  if (packetTr) packetTrace(pt, pkt, rl);

  // Type of MQTT packet. 
  packetIndex = 0;
  switch (pt >> 4) {

  // Remote is trying to connect.
  case MQTT_CONNECT:
    mqttConnect(pt, pkt);
    break;

  // Remote is acknowledging. Not necessary.
  case MQTT_CONNACK:
    //    mqttConnectAck(pt, pkt);
    break;

  // Remote allows these messages.
  case MQTT_SUBSCRIBE:
    mqttSubscribe(pt, pkt);
    break;

  // Some sort of data.
  case MQTT_PUBLISH:
    mqttPublish(pt, pkt, rl);
    break;

  // Publish is complete.
  case MQTT_PUBCOMP:
    mqttPubComp(pt, pkt);
    break;

  // Remote wants to keep us alive.
  case MQTT_PINGREQ:
    mqttPingReq(pt, pkt);
    break;

  // Remote wants to disconnect.
  case MQTT_DISCONNECT:
    mqttDisconnect(pt, pkt);
    break;
    
  // Unknown type but we have a length so skip over it.
  default:
    LOG("Unknown MQTT type %s", mqttTypes[pt >> 4]);
    printf("%s:%s:%d 0x%X stub\n", __FILE__, __func__, __LINE__, pt >> 4);
  }
}


      
// EOF
