/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: packettr.c
 * Author: Jed Marti
 * Description: Trace packets. Write a file with the contents of a packet and crank up the count.
 * Created: Thu Dec  5 09:25:49 2024
 * Revision History:
 *   Sun Dec 29 17:25:06 2024 - Code review.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <sys/time.h>

#include "mqtt.h"

// From the command line.
extern uint8_t packetTr;
extern char    *packetDir;

// The file counter. We use this to generate unique file names for each packet dumped.
uint16_t       packetCount = 0;

// A vector of MQTT packet type names. See packet.c.
extern char *mqttTypes[];


void packetTrace(uint8_t ptype, uint8_t *pkt, int32_t pl)
/*     Dump the current packet in hex and ascii.
 * Parameters:
 *   ptype: MQTT 3.1.1 packet id and DUP/QOS/RETAIN
 *   pkt: Byte of the packet (may not be all).
 *   pl: Number of bytes in the packet.
 * External References:
 *   packetDir: The directory to store the packet in with trailing / if necessary. Must not
 *       be NULL. Can be changed by command line argument -td, default is /tmp/.
 *   packetCount: Count of the packets so they have unique names.
 *   errno: System error for file open.
 *   VERSION, RELEASE, PATCH: constants, indicating what version we're running.
 * Returned values: none
 * Errors detected: Couldn't open output file causes halt.
 */
{
  char    fname[MAXPATHLEN];
  FILE    *fh;
  int32_t i;

  // The file name is dir/count_type-name.tr
  snprintf(fname, MAXPATHLEN, "%s%05d_%s.tr", packetDir, packetCount++,
           mqttTypes[(ptype>>4) & 0x0F]);

  // Create the file.
  if ((fh = fopen(fname, "wb")) == NULL)
  {
    fprintf(stderr, "%s:%s:%d Couldn't open trace file %s because %s\n", __FILE__, __func__,
            __LINE__, fname, strerror(errno));
    exit(-1);
  }

  // Logo with the date and our version.
  fprintf(fh, "%s trace on %s V%d.%02d.%03d\n", mqttTypes[(ptype >> 4) & 0x0F], date(),
          VERSION, RELEASE, PATCH);

  // Write the packet in binary.
  for (i = 0; i < pl; ++i) fprintf(fh, "%c", pkt[i]);
  fprintf(fh, "\n\n");

  // Write out the packet type and length.
  fprintf(fh, " -2: 0x%02X\n -1: 0x%02X\n", ptype, pl);

  // Then dump in a count: hex, and character format.
  for (i = 0; i < pl; ++i)
    if (pkt[i] < 32) fprintf(fh, "%3d: 0x%02X '^%c'\n", i, pkt[i], pkt[i] + 64);
    else fprintf(fh, "%3d: 0x%02X '%c'\n", i, pkt[i], pkt[i]);

  // Done.
  fclose(fh);
}



// EOF
