/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: publish.c
 * Author: Jed Marti
 * Description: This is a real message from a meshtastic device. 
 * Created: Mon Dec  2 06:44:20 2024
 * Revision History:
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#ifdef LINUX
  #include <unistd.h>
#endif
#ifdef WINDOWS
  #include <io.h>
#endif

#include "mqtt.h"
#include "protobuf.h"

// Our host MAC (last 32 bits).
extern uint32_t clientID;

// Network file handle.
extern int32_t FD;

// Noise Level.
extern uint8_t Noise;

// This is the message payload we're concerned with.
uint8_t MSG[MAXMSG];
int32_t MSGLen;

// For tracing publish data from Meshtastic.
int32_t mCnt = 0;
char mttc[1024];

// downlink status.
extern uint8_t downStatus;

// The protoBuf description of a Publish packet.
extern protoBlock protoServiceEnvelope;

// Log files, see log.c, used by LOG mac ro
extern char logMsg[];
extern FILE *LOGfh;

// A list of topics. Items that show up without one of the topics should be ignored. Currently
// we don't bother.
extern Topic *Topics;

/*
uint8_t fcnt = 0;
static void dumpC(char *fname)
//    Dump the PUBLISH packet for debugging.
{
  FILE *fh;
  int32_t i;
  char fn[256];
  snprintf(fn, 256, "%s%02d.c", fname, fcnt++);
  fh = fopen(fn, "wb");
  fprintf(fh, "uint8_t payload[] = {\n");
  for (i = 0; i < MSGLen; ++i)
  {
     fprintf(fh, "   0x%02X, ", MSG[i]);
     if (((i + 1) & 0x0F) == 0) fprintf(fh, "\n");
  }
  fprintf(fh, "};\n");
  fclose(fh);
}
*/



static void collectMeshtastic(uint8_t *pkt, int32_t rl)
/*     Collect the payload and ship it off to the message processing program.
 * Parameters:
 *   pkt: Packet data.
 *   rl: PUBLISH payload size in bytes.
 * External References:
 *   MSG[]: Where to store it.
 *   MSGLen: Number of bytes in it.
 *   Noise: for tracing.
 * Returned values: None.
 * Errors detected: none
 */
{
  printf("collecting %d\n", rl);
  for (MSGLen = 0; MSGLen < rl; ++MSGLen) MSG[MSGLen] = pgetch(pkt);
  //  dumpC("/home/marti/ARES/mqtt/dat/publish");
  if (Noise > 2)
    printf("  %s:%s:%d PUBLISH got %d payload bytes\n", __FILE__, __func__, __LINE__, rl);
}
  



void mqttPublish(uint8_t pt, uint8_t *pkt, int32_t rl)
/*     We have an incoming PUBLISH message. Parse, and send off the stuff to get farmed out.
 * Parameters:
 *   pt: Packet type (ignored).
 *   pkt: The buffer with the data.
 *   rl: The length.
 * External References:
 *   MAXMSG: Maximum message size.
 *   FD: File handle for network I/O.
 *   MSG, MSGLen: Message buffer.
 *   downStatus: Indicates we have enough info from SUBSCRIBE to send messages.
 * Returned values: none
 * Errors Detected: Network write fails.
 */
{
  uint8_t  qos;
  uint16_t tln, i;
  uint16_t pid;
  uint8_t  ack[4];
  char     topic[MAXMSG];
  ssize_t  n;
  uint8_t  cnt;
  Generic  *g;

  // Quality of service.
  qos = (pt >> 1) & 0x03;
  if (Noise > 0)
    printf("%s:%s:%d PUBLISH (%u) QoS %u  0x%02X\n", __FILE__, __func__, __LINE__, rl, qos, pt);

 loop:
  tln = pget16(pkt);
  printf("%s:%s:%d PUBLISH packet length %d\n", __FILE__, __func__, __LINE__, tln);

  // This should never happen. It was usually an indication that a message length was more
  // than one byte.
  if (tln > MAXMSG)
  {
    fprintf(stderr, "%s:%s:%d Bad PUBLISH length %d\n", __FILE__, __func__, __LINE__, tln);
    exit(-1);
  }

  // Print the topic name (or not) but decrement size and skip.
  if (Noise > 1) printf("  %s:%s:%d Topic name (%u):", __FILE__, __func__, __LINE__, tln);
  if (Noise > 1)
  {
    for (i = 0; i < tln; ++i) printf("%c", topic[i] = pgetch(pkt));
    printf("\n");
  }
  else
    for (i = 0; i < tln; ++i) topic[i] = pgetch(pkt);
  rl -= tln;
  topic[i] = 0;
  LOG("PUBLISH topic %s", topic);

  // Only QoS 1 and 2 have a packet id.
  if (qos == 1 || qos == 2)
  {
    pid = pget16(pkt);
    rl -= 2;
    if (Noise > 1) printf("  %s:%s:%d Packet id: %u\n", __FILE__, __func__, __LINE__, pid);
  }

  // Write one of the two responses.
  switch (qos) {
  case 0: break;
  case 1:
    ack[0] = MQTT_PUBACK,
      ack[1] = 2,
      ack[2] = pid >> 8,
      ack[3] = pid & 0xFF;
    if (write(FD, ack, 4) < 0)
    {
      fprintf(stderr, "%s:%s:%d PUBACK fails because %s\n", __FILE__, __func__, __LINE__,
              strerror(errno));
      exit(-1);
    }
    break;

  case 2:
    ack[0] = MQTT_PUBREC,
      ack[1] = 2,
      ack[2] = pid >> 8,
      ack[3] = pid & 0xFF;
    if (write(FD, ack, 4) < 0)
    {
      fprintf(stderr, "%s:%s:%d PUBREC fails because %s\n", __FILE__, __func__, __LINE__,
              strerror(errno));
      exit(-1);
    }
    break;

  default:
    break;
  }

  if (Noise > 1) printf("%s:%s:%d Packet length remaining %d\n", __FILE__, __func__, __LINE__, rl);

  if (rl > 2)
  {
    // Collect the payload bytes and pass off to the mail sender.
    collectMeshtastic(pkt, rl);
    printf("%s:%s:%d collected MSGLen %d\n", __FILE__, __func__, __LINE__, MSGLen);

    // Small messages aren't good.
    if (MSGLen > 5)
    {
      // Pass on to the parser with our base protoBuf.
      g = protoBuf2Generic(MSG, MSGLen, &protoServiceEnvelope);

      // This tells us what to do with the message.
      farmOutPayload(g);

      // We don't need it anymore. Free up the cells. Any values that have been saved were cloned
      // before we get here.
      gc(g);
    }
  }

  // We got our first publish message - we should now have all topics.
  if (downStatus != downUpConnected)
  {
    downStatus = downUpConnected;
  }
  printf("%s:%s:%d PUBLISH finished\n", __FILE__, __func__, __LINE__);
}



/* For tracing output messages.
static int16_t pp = 0;
static void pnputch(uint8_t c) { printf("%03d: 0x%02X '%c'\n", pp++, c, c); }

//#define nputch pnputch
*/

void mqtt2Mesh(uint8_t *msg, int32_t msize)
/*     Publish a message to Meshtastic messages. They go out, they're accepted, but nothing happens.
 */
{
  int32_t sl;
  int32_t i;
  uint8_t slen;
  char    *t;
  char    nt[MAXPATHLEN];

  if ((t = whichIs("ARESMAIL")) == NULL)
  {
    fprintf(stderr, "%s:%s:%d No ARESMAIL channel\n", __FILE__, __func__, __LINE__);
    return;
  }
  //  t = "msh/2/e/ARESMAIL/!da63e1ec";
  //  pp = -2;
  
  // Compute remaining length - topic and message. 
  sl = 2 + strlen(t) + msize + 10;
  printf("%s:%s:%d Sending %d to server\n", __FILE__, __func__, __LINE__, sl);
  
  // MQTT Control packet, no DUP, QOS=0, no retain.
  nputch(0x30);

  // Total packet length is a VARINT.
  if (sl >= 128)
  {
    // Indicates bigger packet (up to 16k).
    nputch((sl & 0x7F) | 0x80);
    nputch(sl >> 7);
  }
  else nputch(sl);

  // Topic length and topic. Remove the + wild card from the end. Lots of warnings about string
  // overflow here which try to ignore.
  nputch(0);
  nputch(strlen(t) - 1);
  slen = (uint8_t) (strlen(t) - 1);
  strncpy(nt, t, slen);
  sprintf(&nt[slen], "!%08x", clientID);
  if (Noise > 1)
    printf("%s:%s:%d publish topic is '%s'\n", __FILE__, __func__, __LINE__, nt);

  // Write out the topic.
  for (i = 0; i < strlen(nt); ++i) nputch(nt[i]);

  // The payload has no length.
  for (i = 0; i < msize; ++i) nputch(msg[i]);

  // Send it.
  nflush();
}


  
// EOF
