/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: scalars.c
 * Author: Jed Marti
 * Description: Protobuf scalars of various sorts.
 * Created: Fri Dec  6 09:21:41 2024
 * Revision History:
 *   Sun Dec 29 21:26:02 2024 - Code review.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>

#include "protobuf.h"

// Table driven stuff.
extern gcFn      gcFns[];
extern displayFn displayFns[];
extern cloneFn   cloneFns[];
extern equalFn   equalFns[];

// Free lists for scalars. These have a 64 bit value.
Generic *FreeScalars = NULL;


Generic *newInt(int64_t v)
/*     Create a new integer. All integers are 64 bits. Conversion from protobuf does the
 *   work.
 * Parameters:
 *   v: Value to box.
 * External References:
 *   FreeScalars: A list of scalar values that can be stored in 64 bits. Use these before
 *     mallocing something new.
 * Returned values: a boxed 64 bit integer.
 * Errors Detected: out of space causes halt.
 */
{
  Int *i;

  if (FreeScalars != NULL)
  {
    i = (Int*) FreeScalars;
    FreeScalars = __next(FreeScalars);
  }
  else if ((i = (Int*) malloc(sizeof(Int))) == NULL)  outOfSpace("Int");

  // Fill in descritor and value.
  __descriptor(i) = DESC_INT,
    __next(i) = NULL,
    __int(i) = v;
  return (Generic*) i;
}



static void freeScalar(Generic *i)
/*     Free an integer by assigning its storage to the FreeStorage list.
 * Parameters:
 *   i: a Generic to free.
 * External References:
 *   FreeScalars: Allocated space for 8 byte values and pairs.
 * Returned values: none
 * Errors detected: none
 */
{
  __next(i) = FreeScalars;
  FreeScalars = i;
}



static void displayInt(FILE *fh, Generic *i)
/*     Display the 64 bit integer.
 * Parameters:
 *   fh: file handle to write to.
 *   i: Boxed integer to write - guaranteed by caller.
 * External References: none
 * Returned values: none
 * Errors detected: none
 */
{
  fprintf(fh, "%ld", __int(i));
}



static Generic *cloneInt(Generic *v)
/*     Clone an integer to make a new one.
 * Parameters:
 *   v: Guaranteed to be an integer.
 * External References: none
 * Returned values: A new integer.
 * Errors detected: none here.
 */
{
  return newInt(__int(v));
}



static uint8_t equalInt(Generic *a, Generic *b)
/*     See if two integers are the same. Part of the Generic equal table driven setup.
 *   See equal.c
 * Parameters:
 *   a, b: Values to check.
 * External References: none
 * Returned values: 0-not equal, 1-equal.
 * Errors detected: none, types checked by caller.
 */
{
  return __int(a) == __int(b);
}



Generic *newDouble(double v)
/*     Create a new boxed double using a free scalar or mallocing a new one.
 * Parameters:
 *   v: Value to box.
 * External References:
 *   FreeScalars: Allocated space for 8 byte values and pairs.
 * Returned values: Boxed double.
 * Errors detected: none
 */
{
  Double *d;

  // Use one of the scalars if they are available.
  if (FreeScalars != NULL)
  {
    d = (Double*) FreeScalars;
    FreeScalars = __next(FreeScalars);
  }

  // No free ones, malloc one.
  else if ((d = (Double*) malloc(sizeof(Double))) == NULL)  outOfSpace("Double");

  // Set the descriptor and fill in the value.
  __descriptor(d) = DESC_DOUBLE,
    __next(d) = NULL,
    __double(d) = v;
  
  return (Generic*) d;
}



void displayDouble(FILE *fh, Generic *d)
/*      Display a double, not all the digits.
 * Parameters:
 *   fh: file handle to write to.
 *   i: Boxed double to write - guaranteed by caller.
 * External References: none
 * Returned values: none
 * Errors detected: none
 */
{
  fprintf(fh, "%g", __double(d));
}



static Generic *cloneDouble(Generic *d)
/*     Clone a double - make a new one.
 * Parameters:
 *   d: A generic double.
 * External References: none
 * Returned values: Boxed double.
 * Errors detected: none
 */
{
  return newDouble(__double(d));
}



static uint8_t equalDouble(Generic *a, Generic *b)
/*     See if two Doubles are the same. Part of the Generic equal table driven setup.
 *   See equal.c. Could probably be done as integers since Doubles are really 64 bit integers
 *   (e.g. Kronecker vs Weierstrass).
 * Parameters:
 *   a, b: Values to check.
 * External References: none
 * Returned values: 0-not equal, 1-equal.
 * Errors detected: none, types checked by caller.
 */
{
  return __double(a) == __double(b);
}



Generic *newPair(Generic *a, Generic *d)
/*     Pairs look like scalars so we use them with the next field is CAR. We use the scalar
 *   freer.
 * Parameters:
 *   a: CAR of pair.
 *   b: CDR of pair.
 * External References:
 *   FreeScalars: Allocated space for 8 byte values and pairs.
 * Returned values: A pair.
 * Errors detected: none
 */
{
  Pair *p;

  // Use scalars.
  if (FreeScalars != NULL)
  {
    p = (Pair*) FreeScalars;
    FreeScalars = __next(FreeScalars);
  }

  // Allocate a new one.
  else if ((p = (Pair*) malloc(sizeof(Pair))) == NULL) outOfSpace("Pair");

  // Fill descriptor and car/cdr.
  __descriptor(p) = DESC_PAIR,
    __next(p) = a,
    __cdr(p) = d;
  return (Generic*) p;
}



static void displayPair(FILE *fh, Generic *p)
/*     Display a pair in list notation.
 * Parameters:
 *   fh: file handle to write to.
 *   i: First pair to show.
 * External References: none
 * Returned values: none
 * Errors detected: none
 */
{
  Generic *q;
  
  fprintf(fh, "(");

  for (q = p; __isPair(q); q = __cdr(q))
  {
    display(fh, __car(q));
    if (!(__isPair(__cdr(q)) || __cdr(q) == NULL))
    {
        fprintf(fh, " . ");
        display(fh, __cdr(q));
        break;
    }
    if (__cdr(q) != NULL) fprintf(fh, " ");
  }
  fprintf(fh, ")");
}



static Generic *clonePair(Generic *p)
/*     Clone a pair and its values to the lowest level.
 * Parameters:
 *   p: A pair (guaranteed by the caller).
 * External References: none
 * Returned values: a copy of the pair to its lowest levels.
 * Errors detected: none
 */
{
  Generic *l = NULL, *le;

  // Scan down the liste.
  for (; __isPair(p); p = __cdr(p))
  {
    if (l == NULL) l = le = newPair(clone(__car(p)), NULL);
    else
    {
     __cdr(le) = newPair(clone(__car(p)), NULL);
     le = __cdr(le);
    }
  }

  // The last element isn't a pair, clone it too.
  if (p != NULL) __cdr(le) = clone(p);
  return l;
}



static uint8_t equalPair(Generic *a, Generic *b)
/*     We have two pairs. Compare them to the lowest level for equality.
 * Parameters:
 *   a, b: Two generics, both guaranteed to be pairs.
 * External References: none
 * Returned values: 0 not equal, 1 equal.
 * Errors detected: none
 */
{
  if (equal(__car(a), __car(b)))
    if (equal(__cdr(a), __cdr(b))) return 1;
  return 0;
}



static void trstring(Generic *s)
/*     Print a string but it may not be zero terminated. Note that the string length is stored
 *   as part of the Generic structure.
 * Parameters:
 *   s: A generic string.
 * External References: none
 * Returned values: none
 * Errors detected: none
 */
{
  int32_t sl, i;
  uint8_t *cs;
  sl = __strlen(s),
    cs = __string(s);
  for (i = 0; i < sl; ++i) printf("%c", cs[i]);
  printf("\n");
}



Generic *assoc(char *what, Generic *l)
/*     Scan a string association list and return the value associated with the string.
 * Parameters:
 *   what: A character string (not a Generic).
 *   l: The association list to search - (("string" . value) ...)
 * External References: none
 * Returned values: The value of the associated string or NULL if not found.
 * Errors detected: none - notice how we had to print the string.
 */
{
  Generic *al, *cal;
  
  if (l == NULL) return NULL;
  for (; l != NULL; l = __cdr(l))
  {
    al = __car(l);
    cal = __car(al);
    if (!__isString(cal))
    {
      fprintf(stderr, "%s:%s:%d assoc bad alist\n", __FILE__, __func__, __LINE__);
      return NULL;
    }
    //    printf("   Working %s == ", what);
    //    trstring(cal); 
    if (strncasecmp(__string(cal), what, (size_t) __next(cal)) == 0) return __cdr(al);
  }
  return NULL;
}



void initScalars(void)
/*     Initialize the GC and display tables. This must get called before any modules setting
 *   their entries in the table.
 * Parameters: none
 * External References:
 *   gcFns[]: table of functions.
 *   displayFns[]: Functions for displaying.
 * Returned values: none
 * Errors detected: none
 */
{
  gcFns[DESC_INT] = freeScalar;
  displayFns[DESC_INT] = displayInt;
  cloneFns[DESC_INT] = cloneInt;
  equalFns[DESC_INT] = equalInt;
  
  gcFns[DESC_DOUBLE] = freeScalar;
  displayFns[DESC_DOUBLE] = displayDouble;
  cloneFns[DESC_DOUBLE] = cloneDouble;
  equalFns[DESC_DOUBLE] = equalDouble;
  
  gcFns[DESC_PAIR] = freeScalar;
  displayFns[DESC_PAIR] = displayPair;
  cloneFns[DESC_PAIR] = clonePair;
  equalFns[DESC_PAIR] = equalPair;
}


// EOF
