/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: strings.c
 * Author: Jed Marti
 * Description: Create and work with dynamic strings.
 * Created: Fri Dec  6 14:31:11 2024
 * Revision History:
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>

#include "protobuf.h"


// Table driven stuff.
extern gcFn      gcFns[];
extern displayFn displayFns[];
extern cloneFn   cloneFns[];
extern equalFn   equalFns[];

// Free lists for strings. The contents are freed but the header isn't.
Generic *FreeStrings = NULL;


Generic *newString(uint8_t *cs, uint32_t sz)
/*     Create a new boxed string duplicating the string. When GC'd, the string gets freed.
 * Parameters:
 *   cs: the string to box.
 * External References:
 *   FreeStrings: A list of those freed.
 * Returned values: the new boxed string.
 */
{
  Generic  *s;
  uint8_t  *u;
  uint32_t i;
  // Use an old one.
  if (FreeStrings != NULL)
  {
    s = FreeStrings;
    FreeStrings = __next(FreeStrings);
  }

  // Malloc a new one.
  else if ((s = (Generic*) malloc(sizeof(String))) == NULL) outOfSpace("String");

  if ((u = (uint8_t*) malloc(sz)) == NULL) outOfSpace("String characters");
  for (i = 0; i < sz; ++i) u[i] = cs[i];
  
  // Set the descriptor and duplicate the string.
  __descriptor(s) = DESC_STRING,
    __next(s) = (void*) (uint64_t) sz,
    __string(s) = u;
  return (Generic*) s;
}



static void freeString(Generic *s)
/*     Free the string by returning the characters to the OS but the box gets put on the
 *   free list.
 * Parameters:
 *   s: String to free, guaranteed by caller.
 * External References:
 *   FreeStrings: A list of those freed.
 * Returned values: none
 * Errors detected: none
 */
{
  if (__string(s) != NULL) free(__string(s));
  __next(s) = FreeStrings;
  FreeStrings = s;
}



static void displayString(FILE *fh, Generic *s)
/*     Display a boxed string in singal quotes.
 * Parameters:
 *   fh: File handle to write to.
 *   s: String to display.
 * External Refernences: none
 * Returned values: none
 * Errors detected: none
 */
{
  uint8_t  *cs;
  uint32_t sz, i;
  
  cs = __string(s);
  sz = (uint64_t) __next(s);
    
  fprintf(fh, "'");
  for (i = 0; i < sz; ++i)
    if (cs[i] < 32) fprintf(fh, "^%c", cs[i] + 65);
    else if (cs[i] >= 128) fprintf(fh, "0x%02x", cs[i]);
    else fprintf(fh, "%c", cs[i]);
  fprintf(fh, "'");
}



static Generic *cloneString(Generic *str)
/*     Clone a string by copying all the values.
 * Parameters:
 *   str: The string to clone.
 * External References: none
 * Returned Values: a new string.
 * Errors detected: none here.
 */
{
  return newString(__string(str), (uint64_t) __next(str));
}



static uint8_t equalString(Generic *a, Generic *b)
/*     For better or worse, strings are mostly case sensitive. But strings are not zero terminated
 *   in our implementation so we have to do this the hard way.
 * Parameters:
 *   a, b: string values to compare. Both are strings guaranteed by caller.
 * External References: none
 * Returned values: 1-all bytes equal, 0-somewhere not equal.
 * Errors detected: none.
 */
{
  if (__strlen(a) != __strlen(b)) return 0;

  if (memcmp(__string(a), __string(b), __strlen(a)) == 0) return 1;
  else return 0;
}



char *string2char(Generic *s)
/*     Convert one of the horrible strings to a zero terminated one. Characters < ' ' become
 *   ^@ and on up. Characters 127 and above become \000 octal. 
 * Paramaters:
 *   s: Generic string (caller tests).
 * External References: none
 * Returned values: a new character string that's 0 terminated. You should free this after use.
 * Errors detected: none.
 */
{
  int32_t sl, i, j;
  uint8_t *c;
  char    m[1024];
  char    oct[6];

  // Degenerate case.
  if (s == NULL) return strdup("(NULL)");
  
  sl = __strlen(s),
    c = __string(s);
  for (j = i = 0; i < sl; ++i)
  {
    if (c[i] < ' ')
    {
      m[j++] = '^';
      m[j++] = c[i] + 64;
    }
    else if (c[i] >= 127)
    {
      sprintf(oct, "%3o", c[i]);
      m[j++] = '\\';
      m[j++] = oct[0];
      m[j++] = oct[1];
      m[j++] = oct[2];
    }
    else m[j++] = c[i];
  }
  m[j] = 0;
  return strdup(m);
}



void initStrings(void)
/*     Initialize the table with the free and display.
 * Parameters: none
 * External References:
 *   gcFns[]: garbage collector table.
 *   displayFns[]: Display functions.
 * Returned values: none
 * Errors detected: none
 */
{
  gcFns[DESC_STRING] = freeString;
  displayFns[DESC_STRING] = displayString;
  cloneFns[DESC_STRING] = cloneString;
  equalFns[DESC_STRING] = equalString;
}



// EOF
