/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: subscribe.c
 * Author: Jed Marti
 * Description: Parse the MQTT subscribe message and acknowledge. The topic does not appear to
 *   be a protoBuf.
 * Created: Tue Dec 10 19:41:13 2024
 * Revision History:
 *   Mon Dec 30 11:56:25 2024 - Code review.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include "mqtt.h"
#include "protobuf.h"

// Network file handle.
extern int32_t FD;

// Noise level. 0-quiet, >0 more and more noise.
extern uint8_t Noise;

// The topic.
uint8_t  *currentTopic = NULL;
uint16_t topicSize = 0;

// Log files, see log.c, used by LOG macro
extern char logMsg[];
extern FILE *LOGfh;

// A list of strings that are current topics. Needed for sending a message. See mqtt.h - just
// a list of strings, not GC'able. Each SUBSCRIBE function sends a topic which looks like:
//     a/b/c/ChannelName/...
Topic *Topics = NULL;


char *whichIs(const char* ba)
/*     Look through the topics and find the one with ARESMAIL in it. This is a crude hack.
 * Parameters:
 *   ba: Topic to look for in a string. We only examine the first 5 characters of the thing we're
 *     looking for which should be enough. 
 * External References:
 *   Topics: The list of topical strings sent by SUBSCRIBE.
 * Returned values: The complete Topic string if found, NULL if not.
 * Errors Detected: none
 */ 
{
  Topic *p;
  char  *fa;
  //  printf("  Looking for %s 0x%0lX\n", ba, (uint64_t) Topics);
  for (p = Topics; p != NULL; p = (Topic*)(p->next))
  {
    //    printf("  %s\n", p->topic);
    fa = strchr(p->topic, ba[0]);
    //    printf("  fa = %s\n", fa);
    if (fa != NULL)
      if (fa[1] == ba[1] && fa[2] == ba[2] && fa[3] == ba[3] && fa[4] == ba[4]) return p->topic;
  }
  return NULL;
}



void clearTopics(void)
/*     We've disconnected, remove the topics and return their storage to the pool.
 * Parameters: none
 * External References:
 *   Topics: The list of topical strings sent by SUBSCRIBE.
 * Returned values: The complete Topic string if found, NULL if not.
 * Errors Detected: none
 */
{
  Topic *p, *q;
  for (p = Topics; p != NULL; p = q)
  {
    q = (Topic*) (p->next);
    free(p->topic);
    free(p);
  }
  Topics = NULL;
}



void mqttSubscribe(uint8_t pt, uint8_t *pkt)
/*     The subscribe MQTT packet. Read such, save the topic for unknown reasons.
 * Parameters:
 *   pt: Packet type (ignored, it's SUBSCRIBE).
 *   pkt: The packet as already collected.
 * External references.
 *   Noise: how much noise to make.
 *   currentTopic: A copy of the current topic.
 *   topicSize: The number of bytes.
 * Returned values: none
 * Errors detected. Write fails.
 */
{
  uint16_t pid;
  uint16_t tsize, i;
  uint8_t  suback[5];
  uint8_t  qos;
  Topic    *t;

  // We got one.
  if (Noise > 0) printf("SUBSCRIBE\n");

  // The protocol ID.
  pid = pget16(pkt);
  tsize = pget16(pkt);

  // Create a buffer for it.
  if ((currentTopic = (uint8_t*) malloc(tsize)) == NULL)
  {
    fprintf(stderr, "%s:%s:%d out of space for currentTopic\n", __FILE__, __func__, __LINE__);
    LOG("SUBSCRIBE out of space, quitting");
    exit(-1);
  }

  // Show the topic.
  if (Noise > 1)
  {
    printf("  %s:%s:%d ID: %u, Size=%u\n", __FILE__, __func__, __LINE__, pid, tsize);
    printf("  %s:%s:%d Topic: ", __FILE__, __func__, __LINE__);
    for (i = 0; i < tsize; ++i) printf("%c", currentTopic[i] = pgetch(pkt));
    printf("\n");
    printf("  %s:%s:%d QoS = 0x%02X\n", __FILE__, __func__, __LINE__, qos = pgetch(pkt));
  }

  // Or ignore it.
  else
    for (i = 0; i < tsize + 1; ++i) currentTopic[i] = pgetch(pkt);

  currentTopic[i] = 0;
  LOG("SUBSCRIBE topic '%s'", currentTopic);

  // Save the topic.
  if ((t = (Topic*) malloc(sizeof(Topic))) == NULL) outOfSpace("Topic");
  t->next = Topics,
    t->topic = currentTopic;
  Topics = t;
  
  currentTopic = NULL;

  // Acknowledge that we got it and QoS.
  suback[0] = 0x90,
    suback[1] = 3,
    suback[2] = pid >> 8,
    suback[3] = pid & 0xFF,
    suback[4] = qos;
  if (write(FD, suback, 5) == -1)
  {
    fprintf(stderr, "%s:%s:%d SUBACK write failed because %s, quitting\n", __FILE__, __func__,
            __LINE__, strerror(errno));
    LOG("SUBACK write failed because %s, quitting.", strerror(errno));
    exit(-1);
  }
}



// EOF
